
! (c) sparslab module name=ainv_n_left
!
! purpose:
!     nonsymmetric left-looking AINV
!     a sample driver
!
!
      use ainv90
!
      implicit none
      integer n,i
      double precision zero
      parameter(zero=0.0d0)
      integer msglvl,msgunit
      integer, allocatable :: ip(:),jp(:)
      double precision, allocatable :: ap(:)
      integer, allocatable :: ipw(:),jpw(:)
      double precision, allocatable :: apw(:)
      integer droptyp,nx
      integer info,ierr
      integer fillw,fillz,imodif,imodifw,nja,size_w,size_z
      double precision drfl,drfl2,diagtol,p1,p2,p3
      integer, allocatable :: ia(:),ja(:)
      double precision, allocatable :: aa(:)
      double precision, allocatable, dimension(:) :: x,y,wr01
      integer, allocatable, dimension(:) :: ib,jb
      double precision, allocatable, dimension(:) :: ab
      integer, allocatable, dimension(:) :: wn01,wn02
!
! start of ainv_n_left
!
!
!  -- input matrix in csr format
!
      p1=1.0d0
      p2=1.0d0
      p3=2.0d0
      nx=25
      n=nx*nx
      nja=(4*2+(nx-2)*(nx-2)*4+4*(nx-2)*3)+n
      allocate(ia(n+1),stat=ierr)
      allocate(ja(nja),aa(nja),stat=ierr)
      call convdiff(nx,p1,p2,p3,ia,ja,aa)
!
!  -- get sample right-hand side
!
!
      allocate(x(n),y(n))
      do i=1,n
        x(i)=dble(i)
      end do
      call mvucsr1(n,ia,ja,aa,x,y)
!
!     %%%%%%%%%%%%%%%%%
!  -- to be set by user
!     %%%%%%%%%%%%%%%%%
!
!
!  -- set the drop tolerance
!  -- for the sainv processs
!
      drfl=0.0d0
      drfl2=0.0d0
!
!     %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
!  -- end of the part to be set by user
!     %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
!
!     %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
!  -- start of defaults -- need not to be changed
!     %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
!
!  -- sizes: initial guess
!
      size_z=max(100,3*(ia(n+1)-1))
      size_w=max(100,3*(ia(n+1)-1))
!
!  -- other parameters without the need to be changed
!
      diagtol=1.1d-16
      droptyp=0
!
!     %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
!  -- end of defaults -- need not to be changed
!     %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
!
!  -- standard level of output
!
      msglvl=2
!
!  -- open unit for output
!
      msgunit=2
      open(unit=msgunit,file='output.out')
!
!  -- call ainv
!
      allocate(wn01(n+1),wn02(n+1),stat=ierr)
      if(ierr.ne.0) then
        ierr=8
        go to 400
      end if
      nja=ia(n+1)-1
      allocate(ib(n+1),jb(nja),ab(nja),stat=ierr)
      if(ierr.ne.0) then
        ierr=8
        go to 400
      end if
      call csr2csc1(n,ia,ja,aa,ib,jb,ab,wn01,wn02,0)
      deallocate(wn01,wn02)
!
      call ainvnl1(2,2,0,n,ia,ja,aa,ib,jb,ab,
     *  ip,jp,ap,size_z,ipw,jpw,apw,size_w,diagtol,
     *  drfl,drfl2,imodif,imodifw,fillz,fillw,info)
!
      if(info .ne.0) go to 400
      deallocate(ia,ja,aa,ib,jb,ab)
!
      write(2,*)' nonzero entries in W'
      write(2,*) apw(1:ipw(n+1)-1)
!
!  -- solve the system (two multiplications)
!
      allocate(wr01(n))
      call mvucsr1(n,ipw,jpw,apw,y,wr01)
      call tvucsr1(n,n,ip,jp,ap,wr01,x)
!
!  -- we have the result in x
!  -- print solution x
!
      write(2,*)' solution x (approximation approx(A^{-1})*b)'
      write(2,*) x(1:n)
!
!  -- print solution x
!
      write(2,*)' solution x'
      write(2,*) x(1:n)
!
!  -- we have the result in y
!
      deallocate(x,y)
!
      close(2)
      stop ' success'
 400  continue
      stop ' some error'
!
      end

