
! (c) sparslab module name=serr
!
! purpose:
!  error output routine.
!
! history:
!   original version for sparslab - tu - 05/10/1997.
!
! parameters :
!  ii  eunit  error unit number.
!  ii  lunit  log unit number.
!  ci  libnam  package name.
!  ci  filnam  routine name.
!  ci  string  output string.
!  ii  err_no  error identifier.
!  ii  err_lev  error level:
!        =-1,0  for warnings; =1 for recoverable errors; =2 for fatal
!        errors.
!
      subroutine serr(eunit,lunit,libnam,filnam,string,err_no,err_lev)
!
! parameters
!
      integer eunit,lunit,err_no,err_lev
      integer len_output_string_max
      parameter(len_output_string_max=100)
!      character*(*) libnam,filnam,string
!      character*(*) libnam,filnam,string
!      character*(len(libnam)) :: libnam
!      character*(len(filnam)) :: filnam
!      character*(len(string)) :: string
      character (len=*) libnam,filnam,string
!      character (len_output_string_max) libnam,filnam,string
!
! functions
!
      integer lenstr0
!
! start of serr
!

      if(eunit.eq.0) then
        write(*,1) '       error no. ',err_no
        write(*,3) ' -- ',string(:lenstr0(string))
        write(*,2) '         found in routine  ',
     *    filnam(:lenstr0(filnam))
        write(*,2) '         called within package  ',
     *    libnam(:lenstr0(libnam))
      else
        write(eunit,1) '       error no. ',err_no
        write(eunit,3) ' -- ',string(:lenstr0(string))
        write(eunit,2) '         found in routine  ',
     *    filnam(:lenstr0(filnam))
        write(eunit,2) '         called within package  ',
     *    libnam(:lenstr0(libnam))
      end if
      if(lunit.ne.eunit) then
        write(lunit,1) '       error no. ',err_no
        write(lunit,3) ' -- ',string(:lenstr0(string))
        write(lunit,2) '         found in routine  ',
     *    filnam(:lenstr0(filnam))
        write(lunit,2) '         called within package  ',
     *    libnam(:lenstr0(libnam))
      end if
      if(err_lev.ge.2) then
!
!    -- fatal error
!
        stop
      end if
!
!  -- formats
!
 1
     * format((a),i7)
 2
     * format((a),2x,(a))
 3
     * format(9x,(a),(a))
!
!  -- return
!
      return
!
! end of serr
!
      end

! (c) sparslab module name=err
!
! purpose:
!  sparslab error handler
!
! history:
!   original version for sparslab - tu - 31/10/2003.
!
! parameters :
!  ii  eunit  error unit number.
!  ci  name  routine name.
!  ci  string  output string.
!  ii  err_no  error identifier.
!  ii  err_lev  error level:
!        =-1,0  for warnings; =1 for recoverable errors; =2 for fatal
!        errors.
!
      subroutine err(eunit,errtyp,name,err_lev)
!
! parameters
!
      use iotypes
      implicit none
      integer eunit,errtyp,err_lev
!      integer len_output_string_max
      character (len=*) name
!
! internals
!
      integer i,ii
      integer iii
      character err_desc*100
!
! functions
!
      integer lenstr,lenstr2
!
! start of err
!
      if(eunit.ne.0) then
        write(eunit,*)
        call errdesc(errtyp,err_desc)
        i=lenstr2(err_desc)
        write(eunit,1) ' error no. ',errtyp
        write(eunit,*) '  -- ',err_desc(:i)
        ii=lenstr(name)
        write(eunit,*) '  found in routine  ',name(:ii)
        write(eunit,*) '  called within the package -- sparslab --'
      end if
      write(*,*)
      call errdesc(errtyp,err_desc)
      i=lenstr2(err_desc)
      write(*,1) ' error no. ',errtyp
      write(*,*) '  -- ',err_desc(:i)
      ii=lenstr(name)
      write(*,*) '  found in routine  ',name(:ii)
      write(*,*) '  called within the package -- sparslab --'
!
      if(err_lev.eq.20) then ! not used tracing at this moment
 100
     *  continue
        iii=lenstr(trace%name)
        write(*,*) '-- called from  ',trace%name(:iii)
        if(trace%level.gt.1) then
          trace=>trace%prev
          go to 100
        end if
      end if
!
      if(err_lev.ge.2) then
!
!    -- fatal error
!
        stop
!
      end if
!
!  -- formats
!
 1
     * format((a),i7)
!
!  -- return
!
      return
!
! end of err
!
      end

! (c) sparslab module name=errdesc
!
! purpose:
!   error / warning  descriptors.
!
! history:
!   original version for sparslab - tu - 31/10/2003.
!
! parameters:
!
      subroutine errdesc(errtyp,err_desc)
!
! parameters
!
      integer errtyp
      integer len_output_string_max
      parameter(len_output_string_max=100)
      character*(len_output_string_max) err_desc
!
! start of errdesc
!
      call stinit(err_desc)
      if(errtyp.eq.0) then
        err_desc=' sucess'
      elseif(errtyp.eq.1) then
        err_desc='dimension nmax too small'
      elseif(errtyp.eq.2) then
        err_desc='dimension mmax too small'
      elseif(errtyp.eq.3) then
        err_desc='matrix is not rectangular'
      elseif(errtyp.eq.4) then
        err_desc='error opening input file'
      elseif(errtyp.eq.5) then
        err_desc='error opening output file'
      elseif(errtyp.eq.6) then
        err_desc='error reading input file'
      elseif(errtyp.eq.7) then
        err_desc='error writing output file'
      elseif(errtyp.eq.8) then
        err_desc='allocation error'
      elseif(errtyp.eq.9) then
        err_desc='deallocation error'
      elseif(errtyp.eq.10) then
        err_desc='error closing file'
      elseif(errtyp.eq.11) then
        err_desc='error reading matrix input type'
      elseif(errtyp.eq.12) then
        err_desc='no space for the diagonal elements in the coox scheme'
      elseif(errtyp.eq.13) then
        err_desc='error reading preconditioning type'
      elseif(errtyp.eq.14) then
        err_desc='error opening matrix definition file slabmatr.df'
      elseif(errtyp.eq.15) then
        err_desc='error reading matrix definition file slabmatr.df'
      elseif(errtyp.eq.16) then
        err_desc='matrix has all elements equal to zero'
      elseif(errtyp.eq.17) then
        err_desc='matrix is not ordered'
      elseif(errtyp.eq.18) then
        err_desc='missing diagonal elements'
      elseif(errtyp.eq.20) then
        err_desc='error while reading data'
      elseif(errtyp.eq.21) then
        err_desc='real working space not sufficient'
      elseif(errtyp.eq.22) then
        err_desc='no space for dynamic data structures'
      elseif(errtyp.eq.23) then
        err_desc='matrix format not supported'
      elseif(errtyp.eq.24) then
        err_desc='matrix is structurally singular'
      elseif(errtyp.eq.25) then
        err_desc='no space for basis columns'
      elseif(errtyp.eq.26) then
        err_desc='matrix used in the linpack is not positive definite'
      elseif(errtyp.eq.27) then
        err_desc='null matrix'
      elseif(errtyp.eq.28) then
        err_desc='unknown norm type'
      elseif(errtyp.eq.29) then
        err_desc='unknown preconditioner type'
      elseif(errtyp.eq.30) then
        err_desc='dimension submax too small'
      elseif(errtyp.eq.31) then
        err_desc='dimension lnzmax too small'
      elseif(errtyp.eq.32) then
        err_desc='not a matrix from mixed-hybrid fem problem'
      elseif(errtyp.eq.33) then
        err_desc='dimension npmax too small'
      elseif(errtyp.eq.34) then
        err_desc='unknown matrix format type'
      elseif(errtyp.eq.35) then
        err_desc='this diag_one is not supported'
      elseif(errtyp.eq.36) then
        err_desc='bad computer setting'
      elseif(errtyp.eq.37) then
        err_desc='bad expl_lev level of expliciteness'
      elseif(errtyp.eq.38) then
        err_desc='min(pmax,mmax) is too small'
      elseif(errtyp.eq.39) then
        err_desc='matrix a is not partially ordered'
      elseif(errtyp.eq.40) then
        err_desc='number of processors greater than anticipated'
      elseif(errtyp.eq.41) then
        err_desc='number of blocks/processes greater than the workspace'
      elseif(errtyp.eq.42) then
        err_desc='unknown element type'
      elseif(errtyp.eq.43) then
        err_desc='unknown boundary condition'
      elseif(errtyp.eq.44) then
        err_desc='corrupted input tg files'
      elseif(errtyp.eq.45) then
        err_desc='modification type not supported'
      elseif(errtyp.eq.46) then
        err_desc='incompatible input data'
      elseif(errtyp.eq.47) then
        err_desc='dimension pmax too small'
      elseif(errtyp.eq.48) then
        err_desc='strange input matrix'
      elseif(errtyp.eq.49) then
        err_desc='this idgtype is not supported'
      elseif(errtyp.eq.50) then
        err_desc='this preconditioner option is not supported'
      elseif(errtyp.eq.51) then
        err_desc='maximum length of an array of variable size exceeded'
      elseif(errtyp.eq.52) then
        err_desc='maximum length of an array of variable size exceeded'
      elseif(errtyp.eq.53) then
        err_desc='shr_prec not supported'
      elseif(errtyp.eq.54) then
        err_desc='nonzero ini_app is not supported in this case'
      elseif(errtyp.eq.55) then
        err_desc='zero row in a matrix'
      elseif(errtyp.eq.56) then
        err_desc='dimension mmax is too small'
      elseif(errtyp.eq.57) then
        err_desc='dimension nzc is too small'
      elseif(errtyp.eq.58) then
        err_desc='the symmetric matrix is not symmetrized'
      elseif(errtyp.eq.61) then
        err_desc='bad input matrix dimension n'
      elseif(errtyp.eq.62) then
        err_desc='bad input number of matrix nonzeros nja (mmax)'
      elseif(errtyp.eq.63) then
        err_desc='small dimension nmax1 for column indices and nonzeros'
      elseif(errtyp.eq.64) then
        err_desc='small dimension for row indices and nonzeros mmax2'
      elseif(errtyp.eq.65) then
        err_desc='small dimension for short vectors nmax1'
      elseif(errtyp.eq.66) then
        err_desc='bad pv_pass'
      elseif(errtyp.eq.67) then
        err_desc='empty column in the initial matrix'
      elseif(errtyp.eq.68) then
        err_desc='empty row in the initial matrix'
      elseif(errtyp.eq.69) then
        err_desc='repeated matrix indices'
      elseif(errtyp.eq.70) then
        err_desc='matrix incorrect'
      elseif(errtyp.eq.201) then
        err_desc='interrupted basis update'
      elseif(errtyp.eq.202) then
        err_desc='not enough space for the preconditioner'
      elseif(errtyp.eq.203) then
        err_desc='lu factorization error'
      elseif(errtyp.eq.601) then
        err_desc='factorial overflow'
      elseif(errtyp.eq.602) then
        err_desc='binomial coefficients overflow'
      elseif(errtyp.eq.801) then
        err_desc='lapack dpotrf error'
      elseif(errtyp.eq.802) then
        err_desc='lapack dpotrf error'
      elseif(errtyp.eq.803) then
        err_desc='lapack dormqr error no.'
      elseif(errtyp.eq.804) then
        err_desc='lapack dgeqrf error no.'
      elseif(errtyp.eq.805) then
        err_desc='lapack dtptrs error no.'
      elseif(errtyp.eq.806) then
        err_desc='lapack dgetrf error no.'
      elseif(errtyp.eq.807) then
        err_desc='lapack dgetrs error no.'
!
!*********************
!****************************************************
! new list of errors
!
!  -- memory faults
!
      elseif(errtyp.eq.1001) then
        err_desc='matrix incorrect -- large indices found'
      elseif(errtyp.eq.1002) then
        err_desc='not a matrix from mixed-hybrid fem problem'
      elseif(errtyp.eq.1010) then
        err_desc='needed integer*8 at least'
!
!  -- format faults
!
      elseif(errtyp.eq.1031) then
        err_desc='matrix format not supported'
      elseif(errtyp.eq.1032) then
        err_desc='diagonal elements assumed'
      elseif(errtyp.eq.1033) then
        err_desc='matrix is not ordered'
      elseif(errtyp.eq.1034) then
        err_desc='matrix file does not contain what expected'
      elseif(errtyp.eq.1041) then
        err_desc='matrix is structurally singular'
      elseif(errtyp.eq.1045) then
        err_desc='matrix is singular and it was modified'
      elseif(errtyp.eq.1046) then
        err_desc='problems with positive definiteness'
      elseif(errtyp.eq.1051) then
        err_desc='matrix format error'
      elseif(errtyp.eq.1052) then
        err_desc='invalid matrix file format'
      elseif(errtyp.eq.1061) then
        err_desc='internal error of dynamic data structures'
      elseif(errtyp.eq.1101) then
        err_desc='allocated space is not sufficient'
      elseif(errtyp.eq.1102) then
        err_desc='space for dynamic data structures is not sufficient'
!
! graph partitioning
!
      elseif(errtyp.eq.1201) then
        err_desc='graph structure is not connected'
      elseif(errtyp.eq.1202) then
        err_desc='number of partitions is not a power of 2'
      elseif(errtyp.eq.1203) then
        err_desc='vertex partitioner is incorrect'
!
!  -- vector errors
!
      elseif(errtyp.eq.1301) then
        err_desc='invalid vector norm type'
!
!  -- direct method errors
!
      elseif(errtyp.eq.1401) then
        err_desc='symbolic factorization failed'
!
!  -- preconditioner computation errors
!
      elseif(errtyp.eq.1501) then
        err_desc='this preconditioning option is not supported'
      elseif(errtyp.eq.1502) then
        err_desc='error reading preconditioner type'
      elseif(errtyp.eq.1511) then
        err_desc='incorrect preconditioner pattern specification'
!
!  -- missing space for hardly predictable space demands
!
      elseif(errtyp.eq.1601) then
        err_desc='missing space to store givens rotations'
      elseif(errtyp.eq.1602) then
        err_desc='missing space to store null space basis'
      elseif(errtyp.eq.1603) then
        err_desc='missing space to store the preconditioner'
      elseif(errtyp.eq.1604) then
        err_desc='missing space to store elementary matrices'
      elseif(errtyp.eq.1605) then
        err_desc='allocation failure'
!
!  -- warnings
!
      elseif(errtyp.eq.1653) then
        err_desc='interrupted update of the preconditioner'
!
!  -- errors in incomplete decompositions
!
!
!  -- warnings
!
      elseif(errtyp.eq.1851) then
        err_desc='missing diagonal entries'
!
!  -- general utilities errors
!
      elseif(errtyp.eq.2001) then
        err_desc='arguments missing or invalid'
      elseif(errtyp.eq.2010) then
        err_desc='syntax error'
!
!  -- i/o errors
!
      elseif(errtyp.eq.2101) then
        err_desc='error opening input file'
      elseif(errtyp.eq.2102) then
        err_desc='error opening output file'
      elseif(errtyp.eq.2103) then
        err_desc='error opening file'
      elseif(errtyp.eq.2104) then
        err_desc='error reading matrix name'
      elseif(errtyp.eq.2105) then
        err_desc='error opening matrix file'
      elseif(errtyp.eq.2106) then
        err_desc='error reading a parameter'
      elseif(errtyp.eq.2107) then
        err_desc='error reading preconditioner type'
      elseif(errtyp.eq.2108) then
        err_desc='error reading right-hand side file'
      elseif(errtyp.eq.2109) then
        err_desc='error reading null-space basis type'
      elseif(errtyp.eq.2110) then
        err_desc='error reading smoother type'
      elseif(errtyp.eq.2111) then
        err_desc='error reading iterative method type'
!
      elseif(errtyp.eq.2201) then
        err_desc='unit does not correspond to an opened file'
!
!  -- sti preprocessor
!
      elseif(errtyp.eq.2501) then
        err_desc='sti syntax error'
!
!  -- text processing routines errors
!
      elseif(errtyp.eq.2651) then
        err_desc='symbol table full -- increase its dimension'
      elseif(errtyp.eq.2652) then
        err_desc='symbol not found in the table'
!
!  -- block methods
!
      elseif(errtyp.eq.3201) then
        err_desc='incorrect static block size'
!
!  -- linpack errors
!
      elseif(errtyp.eq.8001) then
        err_desc='linpack dpofa error'
      elseif(errtyp.eq.8002) then
        err_desc='linpack dpodi error'
!
!  -- lapack errors
!
      elseif(errtyp.eq.8101) then
        err_desc='lapack dpotrf error'
      elseif(errtyp.eq.8102) then
        err_desc='lapack dpotrs error'
!
!  -- sparslab team insufficiencies
!
      elseif(errtyp.eq.9900) then
        err_desc='sorry, this combination of inputs not yet implemented'
      end if
!
!  -- return
!
      return
!
! end of errdesc
!
      end

* (c) sparslab module name=iwset
*
* purpose:
*   set entries of an integer vector to ialpha.
*   unrolled code.
*
* history:
*   original version for sparslab - tu - 14/1/1997
*
* parameters:
*
      subroutine iwset(n,ialpha,ix,incx)
c
c globals
      integer n,ialpha,incx
      integer ix(*)
c
c locals
      integer i,ind,m,mp1
c
c functions
      intrinsic mod
c
c start of iwset
c
      if(n.le.0) return
      if(incx.ne.1) then
c
c incx.ne.1
        ind=1
        if(incx.lt.0) ind=(-n+1)*incx+1
        do i=1,n
          ix(ind)=ialpha
          ind=ind+incx
        end do
        return
      else
c
c incx.eq.1
        m=mod(n,7)
        if(m.ne.0) then
          do i=1,m
            ix(i)=ialpha
          end do
          if(n.lt.7) return
        end if
        mp1=m+1
        do i=mp1,n,7
          ix(i)=ialpha
          ix(i+1)=ialpha
          ix(i+2)=ialpha
          ix(i+3)=ialpha
          ix(i+4)=ialpha
          ix(i+5)=ialpha
          ix(i+6)=ialpha
        end do
        return
      end if
c
c end of iwset
c
      return
c
      end

* (c) sparslab module name=iwset1
*
* purpose:
*   set entries of an integer vector to ialpha.
*   unrolled code.
*
* history:
*   original version for sparslab - tu - 14/1/1997
*
* parameters:
*
      subroutine iwset1(n,ialpha,ix,incx,from)
c
c globals
      integer n,ialpha,incx,from
      integer ix(*)
c
c locals
      integer i,ind,m,mp1
c
c functions
      intrinsic mod
c
c start of iwset1
c
      if(n.le.0) return
      if(incx.ne.1) then
c
c incx.ne.1
        ind=1
        if(incx.lt.0) ind=(-n+1)*incx+1
        do i=1,n
          ix(ind+from-1)=ialpha
          ind=ind+incx
        end do
        return
      else
c
c incx.eq.1
        m=mod(n,7)
        if(m.ne.0) then
          do i=1,m
            ix(i+from-1)=ialpha
          end do
          if(n.lt.7) return
        end if
        mp1=m+1
        do i=mp1,n,7
          ix(i+from-1)=ialpha
          ix(i+1+from-1)=ialpha
          ix(i+2+from-1)=ialpha
          ix(i+3+from-1)=ialpha
          ix(i+4+from-1)=ialpha
          ix(i+5+from-1)=ialpha
          ix(i+6+from-1)=ialpha
        end do
        return
      end if
c
c end of iwset1
c
      return
c
      end

* (c) sparslab module name=ainvrow2
*
* purpose:
*   find a_{indexrc}^t*z^{(indexrc-1)}[i..n]
*   for right-looking stabilized ainv preconditioner.
*
* history:
*   original version for sparslab - tu - 29-03-96.
*   updated - tu - 18/5/1999.
*
* parameters:
*
      subroutine ainvrow2(n,ia,ja,a,ptc,lenc,cnc,h,ptr,lenr,cnr,
     *  wn01,wr01,wn02,wr02,wn03,wn04,ind2,indexrc,dv1,drfl)
c
      integer n
      integer ia(*),ja(*)
      double precision drfl
      double precision a(*)
      integer lenc(*),ptr(*),lenr(*),wn01(*),cnr(*)
      double precision h(*)
      double precision wr01(*),wr02(*)
      integer wn02(*),wn03(*),wn04(*),ptc(*),cnc(*)
c
c internals
c
      integer i,j,k,i1,ii,l
      integer ind3,ind
      integer ind2,indexrc
      integer indk,jstrt,jstop,istrt,istop
      double precision zero,dv1,sum,mi,mifl
      parameter(zero=0.0d0)
      parameter(mi=0.1d0)
c
c start of ainvrow2
c
c  -- find columns intersecting row a_{indexrc}.
c  -- store the indices in wn02(1..ind2).
c  -- indk denotes position of the pivot column indexrc in wn02.
c
c             p_{nit}^{(i-1)} = z_i^{(i-1)}^t*a*z_{nit}^{(i-1)}
c
      n=n
      ind3=0
      mifl=mi*drfl
      istrt=ptc(indexrc)
      istop=ptc(indexrc)+lenc(indexrc)-1
      do i=istrt,istop
        ii=cnc(i)
        jstrt=ia(ii)
        jstop=ia(ii+1)-1
        do j=jstrt,jstop
          k=ja(j)
          l=wn04(k)
          if(l.eq.0) then
            ind3=ind3+1
            wr01(k)=h(i)*a(j)
            wn04(k)=ind3
            wn03(ind3)=k
          else
            wr01(k)=wr01(k)+h(i)*a(j)
          end if
        end do
      end do
c
      ind2=0
      do i=1,ind3
        j=wn03(i)
        jstrt=ptr(j)
        jstop=jstrt+lenr(j)-1
        do j=jstrt,jstop
          k=cnr(j)
          if(wn01(k).eq.0) then
            wn01(k)=j
            ind2=ind2+1
            wn02(ind2)=k
            if(k.eq.indexrc) indk=ind2
          end if
        end do
      end do
c
c  -- loop of dot products between
c  -- columns intersecting row indexrc and
c  -- column z_{indexrc}
c
      do i=1,ind2
        sum=zero
        i1=wn02(i)
        wn01(i1)=0
        jstrt=ptc(i1)
        jstop=jstrt+lenc(i1)-1
        do j=jstrt,jstop
          k=cnc(j)
          if(wr01(k).ne.zero) then
            sum=sum+h(j)*wr01(k)
          end if
        end do
        wr02(i)=sum
      end do
c
      do i=1,ind3
        j=wn03(i)
        wn04(j)=0
        wr01(j)=zero
      end do
c
c  -- get the diagonal pivot dv1
c
      dv1=wr02(indk)
c
c  -- remove the pivot index/entry from the list
c  -- wr02(1..ind2)
c
      wn02(indk)=wn02(ind2)
      wr02(indk)=wr02(ind2)
      ind2=ind2-1
c
c  -- remove small elements
c
      ind=0
      do i=1,ind2
        if(abs(wr02(i)).ge.mifl) then
          ind=ind+1
          wr02(ind)=wr02(i)
          wn02(ind)=wn02(i)
        end if
      end do
      ind2=ind
c
c  -- return
c
      return
c
c end of ainvrow2
c
      end


! (c) sparslab module name=srtcs1
!
! purpose:
!   sort a sparse matrix stored in a cs storage format.
!
! history:
!   original version for sparslab - tu - 01/06/1996.
!
! parameters:
!   ii  n  matrix dimension.
!   ou  ia(n+1)/ja(ia(n+1)-1/a(ia(n+1)-1  updated matrix cs object.
!
      subroutine srtcs1(n,ia,ja,aa)
!
! parameters
!
      implicit none
      integer n,ia(*),ja(*)
      double precision aa(*)
!
! internals
!
      integer i,jlen,jstrt,jstop
!
! start of srtcs1
!
      jstop=ia(1)
      do i=1,n
        jstrt=jstop
        jstop=ia(i+1)
        jlen=jstop-jstrt
        call uxvsr4(jlen,ja(jstrt),aa(jstrt))
      end do
!
!  -- return
!
      return
!
! end of srtcs1
!
      end subroutine srtcs1


! (c) sparslab module name=srtcs1_f90
!
! purpose:
!   sort a sparse matrix stored in a cs storage format.
!
! history:
!   original version for sparslab - tu - 01/06/1996.
!
! parameters:
!   ii  n  matrix dimension.
!   ou  ia(n+1)/ja(ia(n+1)-1/a(ia(n+1)-1  updated matrix cs object.
      subroutine srtcs1_f90(n,ia,ja,a)
!
! parameters
!
      integer n
      integer :: ia(*),ja(*)
      double precision :: a(*)
!
! internals
!
      integer i,jlen,jstrt,jstop
!
! start of srtcs1_f90
!
      jstop=ia(1)
      do i=1,n
        jstrt=jstop
        jstop=ia(i+1)
        jlen=jstop-jstrt
        call srtshair(jlen,ja,a,jstrt,jstrt)
!        call uxvsr4_f90(jlen,ja(jstrt),a(jstrt))
      end do
!
!  -- return
!
      return
!
! end of srtcs1_f90
!
      end

* (c) sparslab module name=garclds
*
* purpose:
*   garbage collection of a matrix in ds* format.
*
* history:
*   original version for sparslab - tu - 13/4/1996.
*
* parameters:
*  ii  n  dimension of the matrix in ds* format.
*  mu  ia(n)/ja(ia(n)+ka(n)-1)/ka(n)/aa(ia(n)+ka(n)-1) updated
*        matrix in ds* format.
*  ii  max_a  size of the vectors ja/aa.
*  ii  ibeg  position of first entry in ia/ja/ka/aa.
*  io  iend  position of last entry in ia/ja/ka/aa.
*  ii  nit  step number (used for modifications of distance parameter).
*  ii  idist  type of the distribution pattern of the ds* format.
*
      subroutine garclds(n,ia,ja,ka,aa,max_a,
     *  ibeg,iend,nit,idist)
c
c parameters
c
      integer n,ibeg,iend,max_a,nit,idist
      integer ia(*),ka(*),ja(*)
      double precision aa(*)
c
c internals
c
      integer i,j,k,l,istrt,isize
      integer izero
      double precision temp
      parameter(izero=0)
c
c start of garclds
c
c  -- compute the minimum necessary size of the compressed output
c
      isize=0
      do i=1,n
        if(ka(i).ne.0) then
          k=ia(i)
          l=ja(k)
          ja(k)=-i
          ia(i)=l
          isize=isize+ka(i)
        else
          ia(i)=ibeg
        end if
      end do
c
c  -- compute distance parameter idist
c
      if(idist.eq.-1) then
c
c  -- -- original case ldist.eq.'a'
c
        idist=max(0,(max_a-1-isize)/(2*(n-nit+1)))
      elseif(idist.gt.0) then
c
c  -- -- original case ldist.eq.'p'
c
        idist=min(max(0,(max_a-1-isize)/(2*(n-nit+1))),idist)
      else
c
c  -- -- original case ldist.eq.'n'
c
        idist=0
      endif
c
c  -- the garbage collection
c
      istrt=ibeg
      i=ibeg
 300
     * continue
      k=ja(i)
c
c  -- -- find a start of some row/column
c
      if(k.lt.0) then
        temp=aa(i)
c
c  -- -- move the row/column
c
        do j=1,ka(-k)-1
          ja(istrt+j)=ja(i+j)
          aa(istrt+j)=aa(i+j)
        end do
c
c  -- -- move the head and redefine the starting pointer
c
        aa(istrt)=temp
        ja(istrt)=ia(-k)
        ia(-k)=istrt
c
c  -- -- redefine starting pointers of both the source and target
c
        istrt=istrt+ka(-k)
        i=i+ka(-k)
c
c  -- -- clear some space (typically not large)
c
        do j=1,min(i-istrt,idist)
          ja(istrt)=0
          istrt=istrt+1
        end do
      else
c
c  -- -- move further
c
        i=i+1
      end if
c
c  -- -- end of coded loop
c
      if(i.lt.iend) go to 300
c
c  -- clear up to the end
c
      call iwset(iend-istrt+1,izero,ja(istrt),1)
c
c  -- redefine end
c
      iend=istrt-1
c
c  -- return
c
      return
c
c end of garclds
c
      end

* (c) sparslab module name=garclds2
*
* purpose:
*   garbage collection of a graph in ds* format.
*
* history:
*   original version for sparslab - tu - 13/4/1996.
*
* parameters:
*  ii  n  dimension of the graph in ds* format.
*  mu  ia(n)/ja(ia(n)+ka(n)-1)/ka(n) updated graph in ds* format.
*  ii  max_a  size of the vectors ja/aa.
*  ii  ibeg  position of first entry in ia/ja/ka/aa.
*  io  iend  position of last entry in ia/ja/ka/aa.
*  ii  idist  type of the distribution pattern of the ds* format.
*
      subroutine garclds2(n,ia,ja,ka,max_a,ibeg,iend,idist)
c
c parameters
c
      integer n,ibeg,iend,max_a,idist
      integer ia(*),ka(*),ja(*)
c
c internals
c
      integer i,j,k,l,istrt,isize
      integer izero
      parameter(izero=0)
c
c start of garclds2
c
c  -- compute the minimum necessary size of the compressed output
c
      isize=0
      do i=1,n
        if(ka(i).ne.0) then
          k=ia(i)
          l=ja(k)
          ja(k)=-i
          ia(i)=l
          isize=isize+ka(i)
        else
          ia(i)=ibeg
        end if
      end do
c
c  -- compute distance parameter idist
c
      if(idist.eq.-1) then
c
c  -- -- original ldist.eq.'a'
c
        idist=max(0,(max_a-1-isize)/(2*(n+1)))
      elseif(idist.gt.0) then
c
c  -- -- original ldist.eq.'p'
c
        idist=min(max(0,(max_a-1-isize)/(2*(n+1))),idist)
      else
c
c  -- -- original case ldist.eq.'n'
c
        idist=0
      endif
c
c  -- the garbage collection
c
      istrt=ibeg
      i=ibeg
 300
     * continue
      k=ja(i)
c
c  -- -- find a start of some row/column
c
      if(k.lt.0) then
c
c  -- -- move the vertex
c
        do j=1,ka(-k)-1
          ja(istrt+j)=ja(i+j)
        end do
c
c  -- -- move the head and redefine the starting pointer
c
        ja(istrt)=ia(-k)
        ia(-k)=istrt
c
c  -- -- redefine starting pointers of both the source and target
c
        istrt=istrt+ka(-k)
        i=i+ka(-k)
c
c  -- -- clear some space (typically not large)
c
        do j=1,min(i-istrt,idist)
          ja(istrt)=0
          istrt=istrt+1
        end do
      else
c
c  -- -- move further
c
        i=i+1
      end if
c
c  -- -- end of coded loop
c
      if(i.lt.iend) go to 300
c
c  -- clear up to the end
c
      call iwset(iend-istrt+1,izero,ja(istrt),1)
c
c  -- redefine end
c
      iend=istrt-1
c
c  -- return
c
      return
c
c end of garclds2
c
      end

* (c) sparslab module name=dwset
*
* purpose:
*   set elements of a double precision vector to a scalar.
*
* history:
*   original version for sparslab - tu - 18/5/1998.
*   blas-like subroutine
*
* parameters:
*
      subroutine dwset(n,alpha,dx,incx)
c
c parameters
c
      integer n,incx
      double precision alpha,dx(*)
c
c internals
c
      integer i,ix,m,mp1
c
c functions
c
      intrinsic mod
c
c start of dwset
c
      if(n.le.0) return
      if(incx.ne.1) then
c
c    -- case incx.ne.1
c
        ix=1
        if(incx.lt.0) ix=(-n+1)*incx+1
        do i=1,n
          dx(ix)=alpha
          ix=ix+incx
        end do
        return
      else
c
c    -- case incx.eq.1
c
        m=mod(n,7)
        if(m.ne.0) then
          do i=1,m
            dx(i)=alpha
          end do
          if(n.lt.7) return
        end if
        mp1=m+1
        do i=mp1,n,7
          dx(i)=alpha
          dx(i+1)=alpha
          dx(i+2)=alpha
          dx(i+3)=alpha
          dx(i+4)=alpha
          dx(i+5)=alpha
          dx(i+6)=alpha
        end do
        return
      end if
c
c end of dwset
c
      end

* (c) sparslab module name=dvcopy_backward
*
* purpose:
*   copy an integer vector in backward order.
*
* history:
*   original version for sparslab - tu - 19/03/2012.
*
* parameters:
*
      subroutine dvcopy_backward(n,dx,dy,from,to)
c
c parameters
c
      integer n,from,to
      double precision dx(*),dy(*)
c
c internals
c
      integer i
c
c functions
c
      intrinsic mod
c
c start of dvcopy_backward
c
      if(n.le.0) return
      do i=n,1,-1
        dy(i+to-1)=dx(i+from-1)
      end do
c
c  -- return
c
      return
c
c end of dvcopy_backward
c
      end

* (c) sparslab module name=iwcopy
*
* purpose:
*   copy an integer vector.
*   unrolled code.
*
* history:
*   original version for sparslab - tu - 14/1/1998.
*   rewritten from diniti.
*
* parameters:
*
      subroutine iwcopy(n,ix,incx,iy,incy)
c
c parameters
c
      integer n,incx,incy,ix(*),iy(*)
c
c locals
      integer i,intx,inty,m,mp1
c
c functions
c
      intrinsic mod
c
c start of iwcopy
c
      if(n.le.0) return
      if(.not.(incx.eq.1.and.incy.eq.1)) then
c
c nonequal increments
        intx=1
        inty=1
        if(incx.lt.0) intx=(-n+1)*incx+1
        if(incy.lt.0) inty=(-n+1)*incy+1
        do i=1,n
          iy(inty)=ix(intx)
          intx=intx+incx
          inty=inty+incy
        end do
      else
c
c increments equal to 1
        m=mod(n,7)
        if(m.ne.0) then
          do i=1,m
            iy(i)=ix(i)
          end do
          if(n.lt.7) return
        end if
        mp1=m+1
        do i=mp1,n,7
          iy(i)=ix(i)
          iy(i+1)=ix(i+1)
          iy(i+2)=ix(i+2)
          iy(i+3)=ix(i+3)
          iy(i+4)=ix(i+4)
          iy(i+5)=ix(i+5)
          iy(i+6)=ix(i+6)
        end do
      end if
c
      return
c
c end of iwcopy
c
      end

* (c) sparslab module name=ivcopy_backward
*
* purpose:
*   copy an integer vector in backward order.
*
* history:
*   original version for sparslab - tu - 19/03/2012.
*
* parameters:
*
      subroutine ivcopy_backward(n,ix,iy,from,to)
c
c parameters
c
      integer n,from,to,ix(*),iy(*)
c
c internals
c
      integer i
c
c functions
c
      intrinsic mod
c
c start of ivcopy_backward
c
      if(n.le.0) return
      do i=n,1,-1
        iy(i+to-1)=ix(i+from-1)
      end do
c
c  -- return
c
      return
c
c end of ivcopy_backward
c
      end

* (c) sparslab module name=lenstr0
*
* purpose:
*  get length of string until character \0
*
* parameters :
*  ci  string  input string.
*  io  lenstr0  length of the string.
*
* history:
*   original version for sparslab - tu - 11/11/1997
*
      integer function lenstr0(string)
c
c parameters
c
      integer len_output_string_max
      parameter(len_output_string_max=100)
      character*(len_output_string_max) string
c
c internals
c
      integer i
c
c start of lenstr0
c
      i=index(string,'\0') - 1
      if(i.eq.-1) then
        i=len(string)
        i=1
      end if
c
      lenstr0=i
c
c  -- return
c
      return
c
c end of lenstr0
c
      end


* (c) sparslab module name=lenstr
*
* purpose:
*  get length of string.
*
* parameters :
*  ci  string  input string.
*  io  lenstr  length of the string.
*
* history:
*   original version for sparslab - tu - 11/11/1997
*
      integer function lenstr(string)
c
c parameters
c
      integer len_output_string_max
      parameter(len_output_string_max=100)
      character*(len_output_string_max) string
c
c internals
c
      integer i
c
c start of lenstr
c
      i=index(string,' ') - 1
      if(i.eq.-1) then
        i=len(string)
        i=1
      end if
c
      lenstr=i
c
c  -- return
c
      return
c
c end of lenstr
c
      end


* (c) sparslab module name=lenstr2
*
* purpose:
*  get length of string.
*  function trying to find the string lengths from back
*  assuming maximal length equal to len_output_string_max.
*  used for string outputs.
*
* parameters :
*  ci  string  input string.
*  io  lenstr2  length of the string.
*
* history:
*   original version for sparslab - tu - 14/9/1998
*
      integer function lenstr2(string)
c
c parameters
c
      integer len_output_string_max
      parameter(len_output_string_max=100)
      character*(len_output_string_max) string
c
c locals
      integer i
c
c start of lenstr2
c
      do i=len_output_string_max,1,-1
        if(string(i:i).ne.' ') then
          lenstr2=i
          return
        end if
      end do
      lenstr2=0
c
      return
c
c end of lenstr2
c
      end


! (c) sparslab module name=mvucsr1
!
! purpose:
!   sparse matrix-vector multiplication y=a*x.
!   csr format.
!
! history:
!   original version for sparslab - tu - 01/05/1996.
!   modified - tu - 29/8/1997
!
! parameters:
!   ii  m  number of matrix rows.
!   mi  ia(m+1)/ja(ia(m+1)-1)/a(ia(m+1)-1)  input matrix csr object.
!   ri  x(n)  input vector.
!   ro  y(m)  output vector y=a*x.
!
      subroutine mvucsr1(m,ia,ja,aa,x,y)
!
! parameters
!
      integer m
      integer ia(*),ja(*)
      double precision aa(*),x(*),y(*)
!
! internals
!
      integer i,j,k,jstrt,jstop
      double precision temp,zero
      parameter(zero=0.0d0)
!
! start of mvucsr1
!
      jstrt=1
      do i=2,m+1
        jstop=ia(i)
        temp=zero
        do j=jstrt,jstop-1
          k=ja(j)
          temp=temp+aa(j)*x(k)
        end do
        y(i-1)=temp
        jstrt=jstop
      end do
!
!  -- return
!
      return
!
! end of mvucsr1
!
      end


* (c) sparslab module name=tvucsr1
*
* purpose:
*   sparse matrix-vector multiplication y=a^t*x; a in csr format.
*   (or sparse matrix-vector multiplication y=a*x; a in csc format.)
*
* history:
*   original version for sparslab - tu - 05-06-96
*   renamed for sparslab - tu - 07-03-96
*   modified - tu - 29/8/1997
*
* parameters:
*   ii  m  number of matrix rows.
*   ii  n  number of matrix columns.
*   oi  ia(m+1)/ja(ia(m+1)-1)/a(ia(m+1)-1)  input matrix csr object.
*   ri  x(m)  input vector.
*   ro  y(n)  output vector y=a*x.
*
c
      subroutine tvucsr1(m,n,ia,ja,a,x,y)
c
c parameters
c
      integer m,n
      integer ia(*),ja(*)
      double precision a(*),x(*),y(*)
c
c locals
c
      integer i,j,k,jstrt,jstop
      double precision temp,zero
      parameter(zero=0.0d0)
c
c start of tvucsr1
c
      call dwset(n,zero,y,1)
      jstrt=1
      do i=2,m+1
        jstop=ia(i)
        temp=x(i-1)
        do j=jstrt,jstop-1
          k=ja(j)
          y(k)=y(k)+a(j)*temp
        end do
        jstrt=jstop
      end do
c
c  -- return
c
      return
c
c end of tvucsr1
c
      end

! (c) sparslab module name=symtr7
!
! purpose :
!  symmetrize a matrix in the css format.
!
! history:
!   original version for sparslab - tu - 24/8/1997
!
! parameters:
!   ii  n dimension of the uncompressed matrix.
!   ii  ncompr dimension of the compressed matrix.
!   ou  ia(n+1)/ja(ia(n+1)-1)/a(ia(n+1)-1)
!         matrix in compressed sparse format.
!         on output it is symmetrized.
!   ia  pd(n)  an auxiliary vector used to store pointers
!         to diagonal elements.
!   ia  wn01(n+1)  an auxiliary vector.
!
      subroutine symtr7(n,ia,ja,a,pd,wn01)
!
! parameters
!
      integer n
      integer ia(*),pd(*),ja(*),wn01(*)
      double precision a(*)
!
! internals
!
      integer i,j,k,ind,istrt,istop,jstrt,jstop
!
! start of symtr7
!
!  -- put (upper) triangular information to an auxiliary array.
!
      jstop=ia(1)
      do i=1,n
        jstrt=jstop
        jstop=ia(i+1)
        wn01(i)=jstop-jstrt
      end do
!
!  -- add (lower) information of the transposed triangle
!  -- to the auxiliary array.
!
      jstop=ia(1)-1
      do i=1,n
        jstrt=jstop+2
        jstop=ia(i+1)-1
        do j=jstrt,jstop
          k=ja(j)
          wn01(k)=wn01(k)+1
        end do
      end do
!
!  -- get pointers of the wide structure
!  -- wn01(i) points at the end of the row i
!
      do i=2,n
        wn01(i)=wn01(i)+wn01(i-1)
      end do
!
!  -- shift the (upper) triangular rows
!  -- set pointers to diagonal elements
!
      pd(1)=1
      jstop=ia(n+1)
      do i=n,1,-1
        ind=wn01(i)
        pd(i+1)=ind+1
        jstrt=jstop-1
        jstop=ia(i)
        do j=jstrt,jstop,-1
          ja(ind)=ja(j)
          a(ind)=a(j)
          ind=ind-1
        end do
      end do
!
!  -- form the transposed (lower) triangular part
!
      istop=ia(1)
      do i=1,n
        istrt=istop
        istop=ia(i+1)
        jstop=wn01(i)
        jstrt=jstop+istrt-istop+2
        do j=jstrt,jstop
          k=ja(j)
          ind=pd(k)
          ja(ind)=i
          a(ind)=a(j)
          pd(k)=ind+1
        end do
      end do
!
!  -- get new pointers
!
      do i=1,n
        ia(i+1)=wn01(i)+1
      end do
!
      return
!
! end of symtr7
!
      end

! (c) sparslab module name=uxvsr4
!
! purpose:
!   ascending shellsort of an integer vector.
!   additional real vector is permuted in the same way.
!
! history:
!   original version for sparslab - tu - 15/01/1998.
!
! parameters:
!  ii  n  input vector length.
!  iu  arrayi(n)  input integer vector to be sorted.
!  ru  arrayr(n)  input real vector to be permuted in the same way.
!
      subroutine uxvsr4(k,arrayi,arrayr)
!
! parameters
!
      implicit none
      integer k
      integer arrayi(*)
      double precision arrayr(*)
!
! internals
!
      integer is,la,lt,ls,lls,i,j,js,khalf,ii
      double precision lb
!
! functions
!
      integer mylog2
!
! start of uxvsr4
!
!  -- nothing to be sorted
!
      if(k.le.1) return
!
!  -- shellsort initialization
!
      khalf=k/2
      ii=mylog2(2*k)
      ls=2**ii-1
!
!  -- shellsort loop
!
      do lt=1,ii
        if(ls.gt.khalf) then
          ls=ls/2
        else
          lls=k-ls
          do 200 i=1,lls
            is=i+ls
            la=arrayi(is)
            lb=arrayr(is)
            j=i
            js=is
 100
     *      continue
            if(la.ge.arrayi(j)) then
              arrayi(js)=la
              arrayr(js)=lb
              go to 200
            else
              arrayi(js)=arrayi(j)
              arrayr(js)=arrayr(j)
              js=j
              j=j-ls
            end if
            if(j.ge.1) go to 100
            arrayi(js)=la
            arrayr(js)=lb
 200
     *  continue
        ls=ls/2
        end if
      end do
!
!  -- return
!
      return
!
! end of uxvsr4
!
      end subroutine uxvsr4



* (c) sparslab module name=srtshair
*
* purpose:
*   ascending shellsort of an integer vector.
*   additional real vector is permuted in the same way.
*
* history:
*   original version for sparslab - tu - 15/01/1998.
*
* parameters:
*  ii  n  input vector length.
*  iu  arrayi(n)  input integer vector to be sorted.
*  ru  arrayr(n)  input real vector to be permuted in the same way.
*
      subroutine srtshair(k,arrayi,arrayr,from,to)
!
! parameters
!
      implicit none
      integer k,from,to
      integer arrayi(*)
      double precision arrayr(*)
!
! internals
!
      integer is,la,lt,ls,lls,i,j,js,khalf,ii
      double precision lb
!
! functions
!
      integer mylog2
!
! start of srtshair
!
!  -- nothing to be sorted
!
      if(k.le.1) return
!
!  -- shellsort initialization
!
      khalf=k/2
      ii=mylog2(2*k)
      ls=2**ii-1
!
!  -- shellsort loop
!
      do lt=1,ii
        if(ls.gt.khalf) then
          ls=ls/2
        else
          lls=k-ls
          do 200 i=1,lls
            is=i+ls
            la=arrayi(is+from-1)
            lb=arrayr(is+to-1)
            j=i
            js=is
 100
     *      continue
            if(la.ge.arrayi(j+from-1)) then
              arrayi(js+from-1)=la
              arrayr(js+to-1)=lb
              go to 200
            else
              arrayi(js+from-1)=arrayi(j+from-1)
              arrayr(js+to-1)=arrayr(j+to-1)
              js=j
              j=j-ls
            end if
            if(j.ge.1) go to 100
            arrayi(js+from-1)=la
            arrayr(js+to-1)=lb
 200
     *  continue
        ls=ls/2
        end if
      end do
!
!  -- return
!
      return
!
! end of srtshair
!
      end


! (c) sparslab module name=mylog2
!
! purpose:
!   upper whole part of logaritm with base two.
!   the name is chosen in order to avoid conflicts
!   with some compilers and their intrinsic funcions.
!
! history:
!   original version for sparslab - tu - 14/02/2001.
!
! parameters:
!
      integer function mylog2(n)
!
! parameters
!
      integer n
!
! internals
!
      integer j,k
!
! start of mylog2
!
      k=0
      j=n
      if(n.gt.1) then
 100
     *  continue
        if(j.eq.1) go to 200
        k=k+1
        j=j/2+mod(j,2)
        go to 100
      end if
 200
     * continue
      mylog2=k
!
!  -- return
!
      return
!
! end of mylog2
!
      end


* (c) sparslab module name=cd2cs
*
* purpose:
*    matrix format transformation routine cd --> cs
*    (some variants of compressed dynamic --> compressed sparse)
*
* history:
*   original version for sparslab - tu - 25/07/2000.
*
      subroutine cd2cs(n,ia,da,ja,la,aa,ib,jb,ab,diag_one)
c
c parameters
c
      integer n,diag_one
      integer ia(*),ja(*),la(*),ib(*),jb(*)
      double precision da(*),aa(*),ab(*)
c
c internals
c
      integer i
c
c start of cd2cs
c
      call cd2cs3(n,ia,la,ja,aa,ib,jb,ab)
      if(diag_one.eq.0) then
        call srtcs1(n,ib,jb,ab)
        do i=1,n
          ab(ib(i+1)-1)=da(i)
        end do
      end if
c
c  -- return
c
      return
c
c end of cd2cs
c
      end

* (c) sparslab module name=cd2cs3
*
* purpose:
*  shrink dynamic data structures in ia/ja/la/aa into
*  ib/jb/ab such that i-th column is moved
*  from the positions ia(i), ..., ia(i)+la(i)-1
*  to the positions ib(i), ..., ib(i+1)-1
*
* history:
*   original version for sparslab - tu - 15/04/1996.
*
* parameters:
*
      subroutine cd2cs3(n,ia,la,ja,aa,ib,jb,ab)
c
c parameters
c
      integer n
      integer ia(*),ja(*),la(*)
      integer ib(*),jb(*)
      double precision ab(*),aa(*)
c
c internals
c
      integer ind,i,j,jstrt,jstop
c
c start of cd2cs3
c
c  -- copy the data
c
      ind=1
      ib(1)=ind
      do i=1,n
        jstrt=ia(i)
        jstop=ia(i)+la(i)-1
        do j=jstrt,jstop
          jb(ind)=ja(j)
          ab(ind)=aa(j)
          ind=ind+1
        end do
        ib(i+1)=ind
      end do
c
c  -- return
c
      return
c
c end of cd2cs3
c
      end



* (C) SPARSLAB MODULE NAME=STINIT
*
* PURPOSE:
*  INITIALIZE AN OUTPUT STRING OF A FIXED SIZE.
*
* HISTORY:
*   ORIGINAL VERSION FOR SPARSLAB - TU - 12/9/1998
*
* PARAMETERS :
*  CU  STRING  A STRING.
*
      SUBROUTINE STINIT(STRING)
C
C GLOBALS
      INTEGER LEN_OUTPUT_STRING_MAX
      PARAMETER(LEN_OUTPUT_STRING_MAX=100)
      CHARACTER*(LEN_OUTPUT_STRING_MAX) STRING
C
C LOCALS
      INTEGER I
C
C START OF STINIT
C
      DO I=1,LEN_OUTPUT_STRING_MAX
        STRING(I:I)=' '
      END DO
C
      RETURN
C
C END OF STINIT
C
      END


! (c) sparslab module name=convdiff
!
! purpose:
!  generate a real square matrix.
!  simple 5pt discretization of 2D convdiff equation.
!
!  -\nabla(u) + 2p_1 u_x + 2 p_2 u_y -p_3 u
!
! Its entries are:
!
!   4-\sigma    \gamma-1  ..... \beta-1 ....
!   -\gamma-1   4-\sigma  ..................
!    .......................................
!   -\beta-1    ............................
!
!
! history:
!   original version for sparslab - tu - 07/05/2008.
!
      subroutine convdiff(nx,p1,p2,p3,ia,ja,aa)
!
! parameters
!
      implicit none
      integer nx,n
      integer :: ia(*),ja(*)
      double precision :: aa(*),p1,p2,p3
!
!
! start of convdiff
!
!  -- convection-diffusion on 2d square grid, dirichlet bc
!
      n=nx*nx
      call grid5pt(nx,n,ia,ja)
      call val5pta(n,ia,ja,aa,nx,p1,p2,p3)
!
!  -- return
!
      return
!
! end of convdiff
!
      end subroutine convdiff



! (c) sparslab module name=grid5pt
!
! purpose:
!   get matrix from the 5pt stencil.
!
! history:
!   original version for sparslab - tu - 18/7/1998.
!
! parameters:
!
      subroutine grid5pt(size,n,ia,ja)
!
      integer size,n
      integer ia(*),ja(*)
!
! internals
!
      integer i,j,k,next
!
! start of grid5pt
!
      next=1
      k=1
      do i=1,size
        do j=1,size
          ia(k)=next
          if(i.ge.2) then
            ja(next)=k-size
            next=next+1
          end if
          if(j.ge.2) then
            ja(next)=k-1
            next=next+1
          end if
          if(j.lt.size) then
            ja(next)=k+1
            next=next+1
          end if
          if(i.lt.size) then
            ja(next)=k+size
            next=next+1
          end if
          k=k+1
        end do
      end do
      ia(k)=next
      n=size*size
!
!  -- return
!
      return
!
! end of grid5pt
!
      end subroutine grid5pt


* (c) sparslab module name=val5pta
*
* purpose:
*   put numerical values into the matrix with 5pt stencil.
*   uses the model:
*           [-u_xx-u_yy+2p_1u_x+2p_2u_y-p_3u] = ...
*
* history:
*   original version for sparslab - tu - 18/7/1998.
*
* parameters:
*
      subroutine val5pta(n,ia,ja,aa,nx,p1,p2,p3)
c
c parameters
c
      implicit none
      integer n
      integer nx
      double precision p1,p2,p3
      integer ia(*),ja(*)
      double precision aa(*)
c
c internals
c
      integer i,j,k,ind,jstrt,jstop,noffdiag,ideg
      double precision hstep,sigma,gamma,beta
      double precision one,four
      parameter(one=1.0d0,four=4.0d0)
c
c start of val5pta
c
      noffdiag=ia(n+1)-1
      ind=noffdiag+n
      do i=noffdiag,1,-1
        ja(ind)=ja(i)
        ind=ind-1
      end do
c
c  -- values
c
      hstep=one/(dble(nx)+one)
      sigma=p3*hstep*hstep
      gamma=p2*hstep
      beta=p1*hstep
c
      ind=1
      jstop=n
      do i=1,n
        ia(i)=ind
        jstrt=jstop+1
        jstop=ia(i+1)-1+n
        ideg=jstop-jstrt+1
        do j=jstrt,jstop
          k=ja(j)
          if(k.lt.i) then
            ja(ind)=ja(j)
              if(k.eq.(i-1)) then
                aa(ind)=gamma-one
              else
                aa(ind)=-beta-one
              end if
            ind=ind+1
          else
            jstrt=j
            exit
          end if
        end do
        ja(ind)=i
        aa(ind)=four-sigma
        if(i.eq.n) jstrt=jstop+1
        ind=ind+1
        do j=jstrt,jstop
          k=ja(j)
          ja(ind)=ja(j)
            if(k.eq.(i+1)) then
              aa(ind)=-gamma-one
            else
              aa(ind)=beta-one
            end if
          ind=ind+1
        end do
      end do
      ia(n+1)=noffdiag+n+1
c
c  -- return
c
      return
c
c end of val5pta
c
      end





* (c) sparslab module name=csr2csc1
*
* purpose:
*  matrix format transformation routine: css,csr --> csc
*                                    or: css,csc --> csr
*
* history:
*   original version for sparslab - tu - 11/02/96
*
* parameters:
*
      subroutine csr2csc1(n,ia,ja,a,im,jm,am,wn01,wn02,mtyp)
c
c parameters
c
      integer n,mtyp
      integer ia(*),ja(*),im(*),jm(*)
      integer wn01(*),wn02(*)
      double precision a(*),am(*)
c
c internals
c
      integer i,j,k,l,jstrt,jstop
      integer izero
      parameter(izero=0)
c
c start of csr2csc1
c
c
c  -- initialization
c
      call iwset(n,izero,wn01,1)
c
c  -- find row counts
c
      do i=1,n
        jstrt=ia(i)
        jstop=ia(i+1)-1
        do j=jstrt,jstop
          k=ja(j)
          wn01(k)=wn01(k)+1
          if(mtyp.eq.1.and.i.ne.k) wn01(i)=wn01(i)+1
        end do
      end do
c
c  -- get partial sums and new pointers
c
      l=1
      wn02(1)=1
      im(1)=1
      do i=2,n+1
        l=l+wn01(i-1)
        wn02(i)=l
        im(i)=l
        wn01(i-1)=0
      end do
c
c  -- rewrite the structure
c
      do i=1,n
        jstrt=ia(i)
        jstop=ia(i+1)-1
        do j=jstrt,jstop
          k=ja(j)
          l=wn02(k)
          jm(l)=i
          am(l)=a(j)
          wn02(k)=wn02(k)+1
          if(mtyp.eq.1.and.i.ne.k) then
            l=wn02(i)
            jm(l)=k
            am(l)=a(j)
            wn02(i)=wn02(i)+1
          end if
        end do
      end do
c
c  -- return
c
      return
c
c end of csr2csc1
c
      end
